import struct, sys

class Buffer(object):
	def __init__(self, data):
		self.data = data
		self.i = 0

	def int8(self):
		self.i += 1
		return struct.unpack('>b', self.data[self.i-1])[0]
	def uint8(self):
		self.i += 1
		return struct.unpack('>B', self.data[self.i-1])[0]
	def int16(self):
		self.i += 2
		return struct.unpack('>h', self.data[self.i-2:self.i])[0]
	def uint16(self):
		self.i += 2
		return struct.unpack('>H', self.data[self.i-2:self.i])[0]
	def int32(self):
		self.i += 4
		return struct.unpack('>l', self.data[self.i-4:self.i])[0]
	def uint32(self):
		self.i += 4
		return struct.unpack('>L', self.data[self.i-4:self.i])[0]
	def double(self):
		self.i += 8
		return struct.unpack('>d', self.data[self.i-8:self.i])[0]
	def string(self, size):
		self.i += size
		return self.data[self.i - size:self.i]
	def databuf(self, size):
		self.i += size
		return map(ord, self.data[self.i - size:self.i])

	def char(self):
		return chr(self.uint8())
	def ptr(self):
		r, q = self.int16(), 0
		if r < 0:
			q = self.int16()
			r = -r
		return q * 32767 + r - 1

	def end(self):
		return self.i == len(self.data)

	def read_field(self, type, ne, var_size):
		if ne == 1:
			assert var_size != -1
			ne = var_size
		if ne != 0:
			return [self.read_field(type, 0, 0) for i in xrange(ne)]

		if type == 'u':
			return self.uint8()
		elif type == 'c':
			return self.int8()
		elif type == 'l':
			v = self.uint8()
			assert v in (0, 1)
			return v == 1
		elif type == 'd':
			return self.int32()
		elif type == 'p':
			return self.ptr()
		elif type == 'f':
			return self.double()
		elif type == 'v':
			return [self.double(), self.double(), self.double()]
		elif type == 'b':
			return [
				self.double(), self.double(), self.double(), 
				self.double(), self.double(), self.double()
			]
		elif type == 'h':
			return [self.double(), self.double(), self.double()]
		else:
			raise Exception('Unknown field type %r' % type)

def load_schema(fn):
	lines = file(fn, 'r').read().strip().split('\n')
	while not lines[0].startswith(': SCHEMA FILE'):
		lines = lines[1:]
	assert lines[-1].startswith('**************** end of schema')
	lines = lines[2:-1]

	node_types = {}
	cur_node = None
	for line in lines:
		if line.find(' ') < line.find(';'): # node type
			type_name, desc, tx_nf_var = line.split('; ')
			type, name = type_name.strip().split(' ')
			type = int(type)
			tx, nf, var = tx_nf_var.strip().split(' ')
			node_types[type] = [name, desc, tx == '1', var == '1', []]
			cur_node = node_types[type][4]
		elif cur_node is not None:
			name, type, tx_nc_ne = line.split('; ')
			tx, nc, ne = tx_nc_ne.strip().split(' ')
			if tx == '1':
				cur_node.append((name, type, int(nc), int(ne)))

	return node_types

def read(fn):
	schema = load_schema('sch_13006.sch')
	
	data = file(fn, 'rb').read()
	if data.startswith('*'):
		eoh = '**END_OF_HEADER*****************************************************************\n'
		cut = data.find(eoh)
		header, data = data[:cut] + eoh, data[cut+len(eoh):]
	assert data[:4] == 'PS\0\0'

	buf = Buffer(data[4:])

	print `buf.string(buf.int16())`
	print `buf.string(buf.int32())`
	user_field_size = buf.int16()
	print 'user_field_size:', user_field_size
	unk = buf.int32()
	assert unk == 0

	types_seen = []
	while not buf.end():
		type = buf.int16()
		if type == 1:
			print 'Terminator!'
			assert buf.uint16() == 1
			assert buf.end()
			break
		if type in schema:
			name, desc, tx, var, fields = schema[type]
			print 'Node %i %s' % (type, name)
			if type not in types_seen: # embedded schema time
				types_seen.append(type)
				nf = buf.uint8()
				if nf != 0xFF:
					new_fields = []
					ptr = 0
					while True:
						c = buf.char()
						if c == 'C':
							new_fields.append(fields[ptr])
							ptr += 1
						elif c == 'D':
							print 'Deleted field:', fields[ptr]
							ptr += 1
						elif c == 'I' or c == 'A':
							if c == 'I':
								print 'Inserting field:', 
							else:
								print 'Appending field:', 
							fname = buf.string(buf.uint8())
							ptr_class = buf.uint16()
							fne = buf.ptr()
							if ptr_class != 0:
								ftype = 'p'
							else:
								ftype = buf.string(buf.uint8())
							if fne == 1:
								xmt_code = buf.uint8()
							new_fields.append((fname, ftype, ptr_class, fne))
							print new_fields[-1]
						elif c == 'Z':
							break
						else:
							raise Exception('Unknown schema character %r' % c)
					schema[type][4] = new_fields
					fields = new_fields
			assert tx
			if var:
				var_size = buf.int32()
				print 'Variable size:', var_size
			else:
				var_size = -1
			node = {}
			index = buf.int16()
			print 'Node index:', index
			for fname, ftype, nc, ne in fields:
				print '\t' + fname, buf.read_field(ftype, ne, var_size)
		else:
			print 'unknown type:', type
			assert False

if __name__=='__main__':
	read(*sys.argv[1:])